from datetime import datetime
from pkg_resources import require
from typing import List, Optional
from urllib.parse import urljoin
from io import BytesIO
import fnmatch

import frontmatter
import requests
import validators
from attr import attrs, attrib
from attr.validators import instance_of, optional
from bs4 import BeautifulSoup
from flask import flash, current_app
from flask_login import UserMixin
from html2text import html2text
from tinydb import Query
from werkzeug.security import generate_password_hash
from werkzeug.datastructures import FileStorage

from archivy import helpers
from archivy.data import create, save_image, valid_image_filename
from archivy.search import add_to_index


# TODO: use this as 'type' field
# class DataobjType(Enum):
#     BOOKMARK = 'bookmark'
#     POCKET_BOOKMARK = 'bookmark imported from pocket'
#     NOTE = 'note'
#     PROCESSED_DATAOBJ = 'bookmark that has been processed'

# TODO: use this as 'type' field
# class DataobjType(Enum):
#     BOOKMARK = 'bookmark'
#     POCKET_BOOKMARK = 'bookmark imported from pocket'
#     NOTE = 'note'
#     PROCESSED_DATAOBJ = 'bookmark that has been processed'
@attrs(kw_only=True)
class DataObj:
    """
    Class that holds a data object (either a note or a bookmark).

    Attributes:

    [Required to pass when creating a new object]

    - **type** -> "note" or "bookmark"

     **Note**:
    - title

    **Bookmark**:

    - url

    [Optional attrs that if passed, will be set by the class]

    - tags
    - content
    - path

    [Handled by the code]

    - id
    - date

    For bookmarks,
    Run `process_bookmark_url()` once you've created it.

    For both types, run `insert()` if you want to create a new file in
    the db with their contents.
    """

    __searchable__ = ["title", "content", "tags"]

    id: Optional[int] = attrib(validator=optional(instance_of(int)), default=None)
    type: str = attrib(validator=instance_of(str))
    title: str = attrib(validator=instance_of(str), default="")
    content: str = attrib(validator=instance_of(str), default="")
    tags: List[str] = attrib(validator=instance_of(list), default=[])
    url: Optional[str] = attrib(validator=optional(instance_of(str)), default=None)
    date: Optional[datetime] = attrib(
        validator=optional(instance_of(datetime)),
        default=None,
    )
    path: str = attrib(validator=instance_of(str), default="")
    fullpath: Optional[str] = attrib(validator=optional(instance_of(str)), default=None)

    def process_bookmark_url(self):
        """Process url to get content for bookmark"""
        if self.type not in ("bookmark", "pocket_bookmark") or not validators.url(
            self.url
        ):
            return None
        selector = None
        for pattern, handler in current_app.config["SCRAPING_PATTERNS"].items():
            if fnmatch.fnmatch(self.url, pattern):
                if type(handler) == str:
                    # if the handler is a string, it's simply a css selector to process the page with
                    selector = handler
                    break
                # otherwise custom user function that overrides archivy behavior
                handler(self)
                return

        try:
            url_request = requests.get(
                self.url,
                headers={"User-agent": f"Archivy/v{require('archivy')[0].version}"},
            )
        except Exception:
            flash(f"Could not retrieve {self.url}\n", "error")
            self.wipe()
            return

        try:
            parsed_html = BeautifulSoup(url_request.text, features="html.parser")
        except Exception:
            flash(f"Could not parse {self.url}\n", "error")
            self.wipe()
            return

        try:
            self.content = self.extract_content(parsed_html, selector)
        except Exception:
            flash(f"Could not extract content from {self.url}\n", "error")
            return

        parsed_title = parsed_html.title
        self.title = parsed_title.string if parsed_title is not None else self.url

    def wipe(self):
        """Resets and invalidates dataobj"""
        self.title = ""
        self.content = ""

    def extract_content(self, beautsoup, selector=None):
        """converts html bookmark url to optimized markdown and saves images"""

        stripped_tags = ["footer", "nav"]
        url = self.url.rstrip("/")

        if selector:
            selected_soup = beautsoup.select(selector)
            # if the custom selector matched, take the first occurrence
            if selected_soup:
                beautsoup = selected_soup[0]
        for tag in stripped_tags:
            if getattr(beautsoup, tag):
                getattr(beautsoup, tag).extract()
        resources = beautsoup.find_all(["a", "img"])
        for tag in resources:
            if tag.name == "a":
                if tag.has_attr("href") and (tag["href"].startswith("/")):
                    tag["href"] = urljoin(url, tag["href"])

                # check it's a normal link and not some sort of image
                # string returns the text content of the tag
                if not tag.string:
                    # delete tag
                    tag.decompose()

            elif tag.name == "img" and tag.has_attr("src"):
                filename = tag["src"].split("/")[-1]
                try:
                    filename = filename[
                        : filename.index("?")
                    ]  # remove query parameters
                except ValueError:
                    pass
                if not tag["src"].startswith("http"):
                    tag["src"] = urljoin(url, tag["src"])
                if current_app.config["SCRAPING_CONF"][
                    "save_images"
                ] and valid_image_filename(filename):
                    image = FileStorage(
                        BytesIO(requests.get(tag["src"]).content), filename, name="file"
                    )
                    saved_to = save_image(image)
                    tag["src"] = "/images/" + saved_to

        res = html2text(str(beautsoup), bodywidth=0)
        return res

    def validate(self):
        """Verifies that the content matches required validation constraints"""
        valid_url = (self.type != "bookmark" or self.type != "pocket_bookmark") or (
            isinstance(self.url, str) and validators.url(self.url)
        )

        valid_title = isinstance(self.title, str) and self.title != ""
        valid_content = self.type not in ("bookmark", "pocket_bookmark") or isinstance(
            self.content, str
        )
        return valid_url and valid_title and valid_content

    def insert(self):
        """Creates a new file with the object's attributes"""
        if self.validate():

            helpers.set_max_id(helpers.get_max_id() + 1)
            self.id = helpers.get_max_id()
            self.date = datetime.now()

            hooks = current_app.config["HOOKS"]

            hooks.before_dataobj_create(self)
            data = {
                "type": self.type,
                "title": str(self.title),
                "date": self.date.strftime("%x").replace("/", "-"),
                "tags": self.tags,
                "id": self.id,
                "path": self.path,
            }
            if self.type == "bookmark" or self.type == "pocket_bookmark":
                data["url"] = self.url

            # convert to markdown file
            dataobj = frontmatter.Post(self.content)
            dataobj.metadata = data
            self.fullpath = str(
                create(
                    frontmatter.dumps(dataobj),
                    f"{self.id}-{dataobj['title']}",
                    path=self.path,
                )
            )

            hooks.on_dataobj_create(self)
            self.index()
            return self.id
        return False

    def index(self):
        return add_to_index(self)

    @classmethod
    def from_md(cls, md_content: str):
        """
        Class method to generate new dataobj from a well formatted markdown string

        Call like this:

        ```python
        Dataobj.from_md(content)

        ```
        """
        data = frontmatter.loads(md_content)
        dataobj = {}
        dataobj["content"] = data.content
        for pair in ["tags", "id", "title", "path"]:
            try:
                dataobj[pair] = data[pair]
            except KeyError:
                # files sometimes get moved temporarily by applications while you edit
                # this can create bugs where the data is not loaded correctly
                # this handles that scenario as validation will simply fail and the event will
                # be ignored
                break

        dataobj["type"] = "processed-dataobj"
        return cls(**dataobj)


@attrs(kw_only=True)
class User(UserMixin):
    """
    Model we use for User that inherits from flask login's
    [`UserMixin`](https://flask-login.readthedocs.io/en/latest/#flask_login.UserMixin)

    Attributes:

    - **username**
    - **password**
    - **is_admin**
    """

    username: str = attrib(validator=instance_of(str))
    password: Optional[str] = attrib(validator=optional(instance_of(str)), default=None)
    is_admin: Optional[bool] = attrib(
        validator=optional(instance_of(bool)), default=None
    )
    id: Optional[int] = attrib(validator=optional(instance_of(int)), default=False)

    def insert(self):
        """Inserts the model from the database"""
        if not self.password:
            return False

        hashed_password = generate_password_hash(self.password)
        db = helpers.get_db()

        if db.search((Query().type == "user") & (Query().username == self.username)):
            return False
        db_user = {
            "username": self.username,
            "hashed_password": hashed_password,
            "is_admin": self.is_admin,
            "type": "user",
        }

        current_app.config["HOOKS"].on_user_create(self)
        return db.insert(db_user)

    @classmethod
    def from_db(cls, db_object):
        """Takes a database object and turns it into a user"""
        username = db_object["username"]
        id = db_object.doc_id

        return cls(username=username, id=id)



import logging
from pathlib import Path
from shutil import which

import elasticsearch
from flask import Flask
from flask_compress import Compress
from flask_login import LoginManager

from archivy import helpers
from archivy.api import api_bp
from archivy.models import User
from archivy.config import Config
from archivy.helpers import load_config, get_elastic_client

app = Flask(__name__)
app.logger.setLevel(logging.INFO)
config = Config()
try:
    # if it exists, load user config
    config.override(load_config(config.INTERNAL_DIR))
except FileNotFoundError:
    pass

app.config.from_object(config)
(Path(app.config["USER_DIR"]) / "data").mkdir(parents=True, exist_ok=True)
(Path(app.config["USER_DIR"]) / "images").mkdir(parents=True, exist_ok=True)

with app.app_context():
    app.config["HOOKS"] = helpers.load_hooks()
    app.config["SCRAPING_PATTERNS"] = helpers.load_scraper()
if app.config["SEARCH_CONF"]["enabled"]:
    with app.app_context():
        search_engines = ["elasticsearch", "ripgrep"]
        es = None
        if (
            "engine" not in app.config["SEARCH_CONF"]
            or app.config["SEARCH_CONF"]["engine"] not in search_engines
        ):
            # try to guess desired search engine if present
            app.logger.warning(
                "Search is enabled but engine option is invalid or absent. Archivy will try to guess preferred search engine."
            )
            app.config["SEARCH_CONF"]["engine"] = "none"

            es = get_elastic_client(error_if_invalid=False)
            if es:
                app.config["SEARCH_CONF"]["engine"] = "elasticsearch"
            else:
                if which("rg"):
                    app.config["SEARCH_CONF"]["engine"] = "ripgrep"
            engine = app.config["SEARCH_CONF"]["engine"]
            if engine == "none":
                app.logger.warning("No working search engine found. Disabling search.")
                app.config["SEARCH_CONF"]["enabled"] = 0
            else:
                app.logger.info(f"Running {engine} installation found.")

        if app.config["SEARCH_CONF"]["engine"] == "elasticsearch":
            es = es or get_elastic_client()
            try:
                es.indices.create(
                    index=app.config["SEARCH_CONF"]["index_name"],
                    body=app.config["SEARCH_CONF"]["es_processing_conf"],
                )
            except elasticsearch.exceptions.RequestError:
                app.logger.info("Elasticsearch index already created")


# login routes / setup
login_manager = LoginManager()
login_manager.login_view = "login"
login_manager.init_app(app)
app.register_blueprint(api_bp, url_prefix="/api")

# compress files
Compress(app)


@login_manager.user_loader
def load_user(user_id):
    db = helpers.get_db()
    res = db.get(doc_id=int(user_id))
    if res and res["type"] == "user":
        return User.from_db(res)
    return None


app.jinja_options["extensions"].append("jinja2.ext.do")


@app.template_filter("pluralize")
def pluralize(number, singular="", plural="s"):
    if number == 1:
        return singular
    else:
        return plural


from archivy import routes  # noqa:



from collections import OrderedDict

import click
from flask import render_template

from archivy import click_web


def index():
    with click.Context(
        click_web.click_root_cmd, info_name=click_web.click_root_cmd.name, parent=None
    ) as ctx:
        return render_template(
            "click_web/show_tree.html",
            ctx=ctx,
            tree=_click_to_tree(ctx, click_web.click_root_cmd),
            title="Plugins",
        )


def _click_to_tree(ctx: click.Context, node: click.BaseCommand, ancestors=[]):
    """
    Convert a click root command to a tree of dicts and lists
    :return: a json like tree
    """
    res_childs = []
    res = OrderedDict()
    res["is_group"] = isinstance(node, click.core.MultiCommand)
    omitted = ["shell", "run", "routes", "create-admin"]
    if res["is_group"]:
        # a group, recurse for e    very child
        subcommand_names = set(click.Group.list_commands(node, ctx))
        children = [
            node.get_command(ctx, key) for key in subcommand_names if key not in omitted
        ]
        # Sort so commands comes before groups
        children = sorted(
            children, key=lambda c: isinstance(c, click.core.MultiCommand)
        )
        for child in children:
            res_childs.append(
                _click_to_tree(
                    ctx,
                    child,
                    ancestors[:]
                    + [
                        node,
                    ],
                )
            )

    res["name"] = node.name

    # Do not include any preformatted block (\b) for the short help.
    res["short_help"] = node.get_short_help_str().split("\b")[0]
    res["help"] = node.help
    path_parts = ancestors + [node]
    res["path"] = "/" + "/".join(p.name for p in path_parts)
    if res_childs:
        res["childs"] = res_childs
    return res



import os
import appdirs


class Config(object):
    """Configuration object for the application"""

    def __init__(self):
        self.SECRET_KEY = os.urandom(32)
        self.PORT = 5000
        self.HOST = "127.0.0.1"
        self.INTERNAL_DIR = appdirs.user_data_dir("archivy")
        self.USER_DIR = self.INTERNAL_DIR
        self.DEFAULT_BOOKMARKS_DIR = ""
        self.SITE_TITLE = "Archivy"
        os.makedirs(self.INTERNAL_DIR, exist_ok=True)

        self.PANDOC_HIGHLIGHT_THEME = "pygments"
        self.SCRAPING_CONF = {"save_images": False}

        self.THEME_CONF = {
            "use_theme_dark": False,
            "use_custom_css": False,
            "custom_css_file": "",
        }

        self.EDITOR_CONF = {
            "settings": {
                "html": False,
                "xhtmlOut": False,
                "breaks": False,
                "linkify": True,
                "typographer": False,
            },
            "plugins": {
                "markdownitFootnote": {},
                "markdownitMark": {},
                "markdownItAnchor": {"permalink": True, "permalinkSymbol": "¶"},
                "markdownItTocDoneRight": {},
            },
        }

        self.SEARCH_CONF = {
            "enabled": 0,
            "url": "http://localhost:9200",
            "index_name": "dataobj",
            "engine": "",
            "es_user": "",
            "es_password": "",
            "es_processing_conf": {
                "settings": {
                    "highlight": {"max_analyzed_offset": 100000000},
                    "analysis": {
                        "analyzer": {
                            "rebuilt_standard": {
                                "stopwords": "_english_",
                                "tokenizer": "standard",
                                "filter": ["lowercase", "kstem", "trim", "unique"],
                            }
                        }
                    },
                },
                "mappings": {
                    "properties": {
                        "title": {
                            "type": "text",
                            "analyzer": "rebuilt_standard",
                            "term_vector": "with_positions_offsets",
                        },
                        "tags": {"type": "text", "analyzer": "rebuilt_standard"},
                        "body": {"type": "text", "analyzer": "rebuilt_standard"},
                    }
                },
            },
        }

    def override(self, user_conf: dict, nested_dict=None):
        """
        This function enables an override of the default configuration with user values.

        Acts smartly so as to only set options already set in the default config.

        - user_conf: current (nested) dictionary of user config key/values
        - nested_dict: reference to the current object that should be modified.
            If none it's just a reference to the current Config itself, otherwise it's a nested dict of the Config
        """
        for k, v in user_conf.items():
            if (nested_dict and not k in nested_dict) or (
                not nested_dict and not hasattr(self, k)
            ):
                # check the key is indeed defined in our defaults
                continue
            curr_default_val = nested_dict[k] if nested_dict else getattr(self, k)
            if type(v) is dict:
                # pass on override to sub configuration dictionary if the current type of value being traversed is dict
                self.override(v, curr_default_val)
            else:
                # otherwise just set
                if type(curr_default_val) == list and type(v) == str:
                    v = v.split(", ")
                if nested_dict:
                    nested_dict[k] = v
                else:
                    setattr(self, k, v)


class BaseHooks:
    """
    Class of methods users can inherit to configure and extend archivy with hooks.


    ## Usage:

    Archivy checks for the presence of a `hooks.py` file in the
    user directory that stores the `data/` directory with your notes and bookmarks.
    This location is usually set during `archivy init`.

    Example `hooks.py` file:

    ```python
    from archivy.config import BaseHooks

    class Hooks(BaseHooks):
        def on_edit(self, dataobj):
            print(f"Edit made to {dataobj.title}")

        def before_dataobj_create(self, dataobj):
            from random import randint
            dataobj.content += f"\\nThis note's random number is {randint(1, 10)}"

        # ...
    ```

    If you have ideas for any other hooks you'd find useful if they were supported,
    please open an [issue](https://github.com/archivy/archivy/issues).
    """

    def on_dataobj_create(self, dataobj):
        """Hook for dataobj creation."""

    def before_dataobj_create(self, dataobj):
        """Hook called immediately before dataobj creation."""

    def on_user_create(self, user):
        """Hook called after a new user is created."""

    def on_edit(self, dataobj):
        """Hook called whenever a user edits through the web interface or the API."""



from pathlib import Path
from shutil import which
from subprocess import run, PIPE

from flask import current_app

from archivy.helpers import get_elastic_client

RG_REGEX_ARG = "-e"
RG_MISC_ARGS = "-ilt"  # i -> case insensitive and l -> only output filenames
RG_FILETYPE = "md"


def add_to_index(model):
    """
    Adds dataobj to given index. If object of given id already exists, it will be updated.

    Params:

    - **index** - String of the ES Index. Archivy uses `dataobj` by default.
    - **model** - Instance of `archivy.models.Dataobj`, the object you want to index.
    """
    es = get_elastic_client()
    if not es:
        return
    payload = {}
    for field in model.__searchable__:
        payload[field] = getattr(model, field)
    es.index(
        index=current_app.config["SEARCH_CONF"]["index_name"], id=model.id, body=payload
    )
    return True


def remove_from_index(dataobj_id):
    """Removes object of given id"""
    es = get_elastic_client()
    if not es:
        return
    es.delete(index=current_app.config["SEARCH_CONF"]["index_name"], id=dataobj_id)


def query_es_index(query, strict=False):
    """
    Returns search results for your given query

    Specify strict=True if you want only exact result (in case you're using ES.
    """
    es = get_elastic_client()
    if not es:
        return []
    search = es.search(
        index=current_app.config["SEARCH_CONF"]["index_name"],
        body={
            "query": {
                "multi_match": {
                    "query": query,
                    "fields": ["*"],
                    "analyzer": "rebuilt_standard",
                }
            },
            "highlight": {
                "fragment_size": 0,
                "fields": {
                    "content": {
                        "pre_tags": "==",
                        "post_tags": "==",
                    }
                },
            },
        },
    )

    hits = []
    for hit in search["hits"]["hits"]:
        formatted_hit = {"id": hit["_id"], "title": hit["_source"]["title"]}
        if "highlight" in hit:
            formatted_hit["highlight"] = hit["highlight"]["content"]
            reformatted_match = " ".join(formatted_hit["highlight"]).replace("==", "")
            if strict and not (query in reformatted_match):
                continue
        hits.append(formatted_hit)
    return hits


def query_ripgrep(query):
    """Uses ripgrep to search data with a simpler setup than ES"""

    from archivy.data import get_data_dir

    if current_app.config["SEARCH_CONF"]["engine"] != "ripgrep" or not which("rg"):
        return None

    rg_cmd = ["rg", RG_MISC_ARGS, RG_FILETYPE, RG_REGEX_ARG, query, str(get_data_dir())]
    rg = run(rg_cmd, stdout=PIPE, stderr=PIPE, timeout=60)
    file_paths = [Path(p.decode()).parts[-1] for p in rg.stdout.splitlines()]

    # don't open file just find info from filename for speed
    hits = []
    for filename in file_paths:
        parsed = filename.replace(".md", "").split("-")
        hits.append({"id": int(parsed[0]), "title": "-".join(parsed[1:])})
    return hits


def search(query, strict=False):
    """
    Wrapper to search methods for different engines.

    If using ES, specify strict=True if you only want results that strictly match the query, without parsing / tokenization.
    """
    if current_app.config["SEARCH_CONF"]["engine"] == "elasticsearch":
        return query_es_index(query, strict=strict)
    elif current_app.config["SEARCH_CONF"]["engine"] == "ripgrep":
        return query_ripgrep(query)



import tempfile
from pathlib import Path

import click
import jinja2
from flask import Blueprint

from archivy.click_web.resources import cmd_exec, cmd_form, index

jinja_env = jinja2.Environment(extensions=["jinja2.ext.do"])

"The full path to the click script file to execute."
script_file = None
"The click root command to serve"
click_root_cmd = None


def _get_output_folder():
    _output_folder = Path(tempfile.gettempdir()) / "click-web"
    if not _output_folder.exists():
        _output_folder.mkdir()
    return _output_folder


"Where to place result files for download"
OUTPUT_FOLDER = str(_get_output_folder())

_flask_app = None
logger = None


def create_click_web_app(module, command: click.BaseCommand, flask_app):
    """
    Create a Flask app that wraps a click command. (Call once)

    :param module: the module that contains the click command,
    needed to get the path to the script.
    :param command: The actual click root command, needed
    to be able to read the command tree and arguments
    in order to generate the index page and the html forms
    """
    global _flask_app, logger
    assert _flask_app is None, "Flask App already created."

    _register(module, command)

    _flask_app = flask_app

    # add the "do" extension needed by our jinja templates
    _flask_app.jinja_env.add_extension("jinja2.ext.do")

    _flask_app.add_url_rule("/plugins", "cli_index", index.index)
    _flask_app.add_url_rule(
        "/cli/<path:command_path>", "command", cmd_form.get_form_for
    )
    _flask_app.add_url_rule(
        "/cli/<path:command_path>", "command_execute", cmd_exec.exec, methods=["POST"]
    )

    _flask_app.logger.info(f"OUTPUT_FOLDER: {OUTPUT_FOLDER}")
    results_blueprint = Blueprint(
        "results",
        __name__,
        static_url_path="/static/results",
        static_folder=OUTPUT_FOLDER,
    )
    _flask_app.register_blueprint(results_blueprint)

    logger = _flask_app.logger

    return _flask_app


def _register(module, command: click.BaseCommand):
    """
    :param module: the module that contains the command, needed to get the path to the script.
    :param command: The actual click root command,
                    needed to be able to read the command tree and arguments
                    in order to generate the index page and the html forms
    """
    global click_root_cmd, script_file
    script_file = str(Path(module.__file__).absolute())
    click_root_cmd = command



import os
import shutil
import subprocess
import sys
import tempfile
import traceback
from pathlib import Path
from typing import List

from flask import Response, request
from werkzeug.utils import secure_filename

from archivy import click_web

from .input_fields import FieldId

logger = None


def exec(command_path):
    """
    Execute the command and stream the output from it as response
    :param command_path:
    """
    command_path = "cli/" + command_path
    global logger
    logger = click_web.logger

    omitted = ["shell", "run", "routes", "create-admin"]
    root_command, *commands = command_path.split("/")
    cmd = ["archivy"]
    req_to_args = RequestToCommandArgs()
    # root command_index should not add a command
    cmd.extend(req_to_args.command_args(0))
    for i, command in enumerate(commands):
        if command in omitted:
            return Response(status=400)
        cmd.append(command)
        cmd.extend(req_to_args.command_args(i + 1))

    def _generate_output():
        yield _create_cmd_header(commands)
        try:
            yield from _run_script_and_generate_stream(req_to_args, cmd)
        except Exception as e:
            # exited prematurely, show the error to user
            yield f"\nERROR: Got exception when reading output from script: {type(e)}\n"
            yield traceback.format_exc()
            raise

    return Response(_generate_output(), mimetype="text/plain")


def _run_script_and_generate_stream(
    req_to_args: "RequestToCommandArgs", cmd: List[str]
):
    """
    Execute the command the via Popen and yield output
    """
    logger.info("Executing archivy command")
    if not os.environ.get("PYTHONIOENCODING"):
        # Fix unicode on windows
        os.environ["PYTHONIOENCODING"] = "UTF-8"

    process = subprocess.Popen(
        cmd, shell=False, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    )
    logger.info("script running Pid: %d", process.pid)

    encoding = sys.getdefaultencoding()
    with process.stdout:
        for line in iter(process.stdout.readline, b""):
            yield line.decode(encoding)

    process.wait()  # wait for the subprocess to exit
    logger.info("script finished Pid: %d", process.pid)
    for fi in req_to_args.field_infos:
        fi.after_script_executed()


def _create_cmd_header(commands: List[str]):
    """
    Generate a command header.
    Note:
        here we always allow to generate HTML as long as we have it between CLICK-WEB comments.
        This way the JS frontend can insert it in the correct place in the DOM.
    """

    def generate():
        yield "<!-- CLICK_WEB START HEADER -->"
        yield '<div class="command-line">Executing: {}</div>'.format("/".join(commands))
        yield "<!-- CLICK_WEB END HEADER -->"

    # important yield this block as one string so it pushed to client in one go.
    # so the whole block can be treated as html.
    html_str = "\n".join(generate())
    return html_str


def _create_result_footer(req_to_args: "RequestToCommandArgs"):
    """
    Generate a footer.
    Note:
        here we always allow to generate HTML as long as we have it between CLICK-WEB comments.
        This way the JS frontend can insert it in the correct place in the DOM.
    """
    to_download = [
        fi
        for fi in req_to_args.field_infos
        if fi.generate_download_link and fi.link_name
    ]
    # important yield this block as one string so it pushed to client in one go.
    # This is so the whole block can be treated as html if JS frontend.
    lines = []
    lines.append("<!-- CLICK_WEB START FOOTER -->")
    if to_download:
        lines.append("<b>Result files:</b><br>")
        for fi in to_download:
            lines.append("<ul> ")
            lines.append(f"<li>{_get_download_link(fi)}<br>")
            lines.append("</ul>")

    else:
        lines.append("<b>DONE</b>")
    lines.append("<!-- CLICK_WEB END FOOTER -->")
    html_str = "\n".join(lines)
    yield html_str


def _get_download_link(field_info):
    """Hack as url_for need request context"""

    rel_file_path = Path(field_info.file_path).relative_to(click_web.OUTPUT_FOLDER)
    uri = f"/static/results/{rel_file_path.as_posix()}"
    return f'<a href="{uri}">{field_info.link_name}</a>'


class RequestToCommandArgs:
    def __init__(self):
        field_infos = [
            FieldInfo.factory(key)
            for key in list(request.form.keys()) + list(request.files.keys())
        ]
        # important to sort them so they will be in expected order on command line
        self.field_infos = list(sorted(field_infos))

    def command_args(self, command_index) -> List[str]:
        """
        Convert the post request into a list of command line arguments

        :param command_index: (int) the index for the command to get arguments for.
        :return: list of command line arguments for command at that cmd_index
        """
        args = []

        # only include relevant fields for this command index
        commands_field_infos = [
            fi for fi in self.field_infos if fi.param.command_index == command_index
        ]
        commands_field_infos = sorted(commands_field_infos)

        for fi in commands_field_infos:

            # must be called mostly for saving and preparing file output.
            fi.before_script_execute()

            if fi.cmd_opt.startswith("--"):
                # it's an option
                args.extend(self._process_option(fi))

            else:
                # argument(s)
                if isinstance(fi, FieldFileInfo):
                    # it's a file, append the written temp file path
                    # TODO: does file upload support multiple keys? In that case support it.
                    args.append(fi.file_path)
                else:
                    arg_values = request.form.getlist(fi.key)
                    has_values = bool("".join(arg_values))
                    if has_values:
                        if fi.param.nargs == -1:
                            # Variadic argument, in html form each argument
                            # is a separate line in a textarea.
                            # treat each line we get from text area as a separate argument.
                            for value in arg_values:
                                values = value.splitlines()
                                logger.info(
                                    f'variadic arguments, split into: "{values}"'
                                )
                                args.extend(values)
                        else:
                            logger.info(f'arg_value: "{arg_values}"')
                            args.extend(arg_values)
        return args

    def _process_option(self, field_info):
        vals = request.form.getlist(field_info.key)
        if field_info.is_file:
            if field_info.link_name:
                # it's a file, append the file path
                yield field_info.cmd_opt
                yield field_info.file_path
        elif field_info.param.param_type == "flag":
            # To work with flag that is default True
            # a hidden field with same name is also sent by form.
            # This is to detect if checkbox was not checked as then
            # we will get the field anyway with the "off flag" as value.
            if len(vals) == 1:
                off_flag = vals[0]
                flag_on_cmd_line = off_flag
            else:
                # we got both off and on flags, checkbox is checked.
                on_flag = vals[1]
                flag_on_cmd_line = on_flag

            yield flag_on_cmd_line
        elif "".join(vals):
            # opt with value, if option was given multiple times get the values for each.
            # flag options should always be set if we get them
            # for normal options they must have a non empty value
            yield field_info.cmd_opt
            for val in vals:
                if val:
                    yield val
        else:
            # option with empty values, should not be added to command line.
            pass


class FieldInfo:
    """
    Extract information from the encoded form input field name
    the parts:
        [command_index].[opt_or_arg_index].[click_type].[html_input_type].[opt_or_arg_name]
    e.g.
        "0.0.option.text.text.--an-option"
        "0.1.argument.file[rb].text.an-argument"
    """

    @staticmethod
    def factory(key):
        field_id = FieldId.from_string(key)
        is_file = field_id.click_type.startswith("file")
        is_path = field_id.click_type.startswith("path")
        is_uploaded = key in request.files
        if is_file:
            if is_uploaded:
                field_info = FieldFileInfo(field_id)
            else:
                field_info = FieldOutFileInfo(field_id)
        elif is_path:
            if is_uploaded:
                field_info = FieldPathInfo(field_id)
            else:
                field_info = FieldPathOutInfo(field_id)
        else:
            field_info = FieldInfo(field_id)
        return field_info

    def __init__(self, param: FieldId):
        self.param = param
        self.key = param.key

        "Type of option (file, text)"
        self.is_file = self.param.click_type.startswith("file")

        "The actual command line option (--debug)"
        self.cmd_opt = param.name

        self.generate_download_link = False

    def before_script_execute(self):
        pass

    def after_script_executed(self):
        pass

    def __str__(self):
        return str(self.param)

    def __lt__(self, other):
        "Make class sortable"
        return (self.param.command_index, self.param.param_index) < (
            other.param.command_index,
            other.param.param_index,
        )

    def __eq__(self, other):
        return self.key == other.key


class FieldFileInfo(FieldInfo):
    """
    Use for processing input fields of file type.
    Saves the posted data to a temp file.
    """

    "temp dir is on class in order to be uniqe for each request"
    _temp_dir = None

    def __init__(self, fimeta):
        super().__init__(fimeta)
        # Extract the file mode that is in the type e.g file[rw]
        self.mode = self.param.click_type.split("[")[1][:-1]
        self.generate_download_link = True if "w" in self.mode else False
        self.link_name = f"{self.cmd_opt}.out"

        logger.info(f"File mode for {self.key} is {self.mode}")

    def before_script_execute(self):
        self.save()

    @classmethod
    def temp_dir(cls):
        if not cls._temp_dir:
            cls._temp_dir = tempfile.mkdtemp(dir=click_web.OUTPUT_FOLDER)
        logger.info(f"Temp dir: {cls._temp_dir}")
        return cls._temp_dir

    def save(self):
        logger.info("Saving...")

        logger.info("field value is a file! %s", self.key)
        file = request.files[self.key]
        # if user does not select file, browser also
        # submit a empty part without filename
        if file.filename == "":
            raise ValueError("No selected file")
        elif file and file.filename:
            filename = secure_filename(file.filename)
            name, suffix = os.path.splitext(filename)

            fd, filename = tempfile.mkstemp(
                dir=self.temp_dir(), prefix=name, suffix=suffix
            )
            self.file_path = filename
            logger.info(f"Saving {self.key} to {filename}")
            file.save(filename)

    def __str__(self):

        res = [super().__str__()]
        res.append(f"file_path: {self.file_path}")
        return ", ".join(res)


class FieldOutFileInfo(FieldFileInfo):
    """
    Used when file option is just for output and form posted it as hidden or text field.
    Just create a empty temp file to give it's path to command.
    """

    def __init__(self, fimeta):
        super().__init__(fimeta)
        if self.param.form_type == "text":
            self.link_name = request.form[self.key]
            # set the postfix to name name provided from form
            # this way it will at least have the same extension when downloaded
            self.file_suffix = request.form[self.key]
        else:
            # hidden no preferred file name can be provided by user
            self.file_suffix = ".out"

    def save(self):
        name = secure_filename(self.key)

        filename = tempfile.mkstemp(
            dir=self.temp_dir(), prefix=name, suffix=self.file_suffix
        )
        logger.info(f"Creating empty file for {self.key} as {filename}")
        self.file_path = filename


class FieldPathInfo(FieldFileInfo):
    """
    Use for processing input fields of path type.
    Extracts the posted data to a temp folder.
    When script finished zip that folder and provide download link to zip file.
    """

    def save(self):
        super().save()
        zip_extract_dir = tempfile.mkdtemp(dir=self.temp_dir())

        logger.info(f"Extracting: {self.file_path} to {zip_extract_dir}")
        shutil.unpack_archive(self.file_path, zip_extract_dir, "zip")
        self.file_path = zip_extract_dir

    def after_script_executed(self):
        super().after_script_executed()
        fd, filename = tempfile.mkstemp(dir=self.temp_dir(), prefix=self.key)
        folder_path = self.file_path
        self.file_path = filename

        logger.info(f"Zipping {self.key} to {filename}")
        self.file_path = shutil.make_archive(self.file_path, "zip", folder_path)
        logger.info(f"Zip file created {self.file_path}")
        self.generate_download_link = True


class FieldPathOutInfo(FieldOutFileInfo):
    """
    Use for processing output fields of path type.
    Create a folder and use as path to script.
    When script finished zip that folder and provide download link to zip file.
    """

    def save(self):
        super().save()
        self.file_path = tempfile.mkdtemp(dir=self.temp_dir())

    def after_script_executed(self):
        super().after_script_executed()
        fd, filename = tempfile.mkstemp(dir=self.temp_dir(), prefix=self.key)
        folder_path = self.file_path
        self.file_path = filename
        logger.info(f"Zipping {self.key} to {filename}")
        self.file_path = shutil.make_archive(self.file_path, "zip", folder_path)
        logger.info(f"Zip file created {self.file_path}")
        self.generate_download_link = True



import platform
import subprocess
import os
from pathlib import Path
import shutil

import frontmatter
from flask import current_app
from werkzeug.utils import secure_filename
from werkzeug.datastructures import FileStorage

from archivy.search import remove_from_index


# FIXME: ugly hack to make sure the app path is evaluated at the right time
def get_data_dir():
    """Returns the directory where dataobjs are stored"""
    return Path(current_app.config["USER_DIR"]) / "data"


def is_relative_to(sub_path, parent):
    """Implement pathlib `is_relative_to` only available in python 3.9"""
    try:
        sub_path.resolve().relative_to(parent)
        return True
    except ValueError:
        return False


class Directory:
    """Tree like file-structure used to build file navigation in Archiv"""

    def __init__(self, name):
        self.name = name
        self.child_files = []
        self.child_dirs = {}


FILE_GLOB = "[0-9]*-*.md"


def get_by_id(dataobj_id):
    """Returns filename of dataobj of given id"""
    results = list(get_data_dir().rglob(f"{dataobj_id}-*.md"))
    return results[0] if results else None


def get_items(collections=[], path="", structured=True, json_format=False):
    """
    Gets all dataobjs.

    Parameters:

    - **collections** - filter dataobj by type, eg. bookmark / note
    - **path** - filter by path
    - **structured: if set to True, will return a Directory object, otherwise
      data will just be returned as a list of dataobjs
    - **json_format**: boolean value used internally to pre-process dataobjs
      to send back a json response.
    """
    datacont = Directory(path or "root") if structured else []
    data_dir = get_data_dir()
    root_dir = data_dir / path
    if not is_relative_to(root_dir, data_dir) or not root_dir.exists():
        raise FileNotFoundError
    if structured:
        for filepath in root_dir.rglob("*"):
            current_path = filepath.relative_to(root_dir)
            current_dir = datacont

            # iterate through parent directories
            for segment in current_path.parts[:-1]:
                # directory has not been saved in tree yet
                if segment not in current_dir.child_dirs:
                    current_dir.child_dirs[segment] = Directory(segment)
                current_dir = current_dir.child_dirs[segment]

            # handle last part of current_path
            last_seg = current_path.parts[-1]
            if filepath.is_dir():
                if last_seg not in current_dir.child_dirs:
                    current_dir.child_dirs[last_seg] = Directory(last_seg)
                current_dir = current_dir.child_dirs[last_seg]
            elif last_seg.endswith(".md"):
                data = frontmatter.load(filepath)
                current_dir.child_files.append(data)
    else:
        for filepath in root_dir.rglob("*.md"):
            data = frontmatter.load(filepath)
            data["fullpath"] = str(filepath.parent.relative_to(root_dir))
            if len(collections) == 0 or any(
                [collection == data["type"] for collection in collections]
            ):
                if json_format:
                    dict_dataobj = data.__dict__
                    # remove unnecessary yaml handler
                    dict_dataobj.pop("handler")
                    datacont.append(dict_dataobj)
                else:
                    datacont.append(data)
    return datacont


def create(contents, title, path=""):
    """
    Helper method to save a new dataobj onto the filesystem.

    Parameters:

    - **contents**: md file contents
    - **title** - title used for filename
    - **path**
    """
    filename = secure_filename(title)
    data_dir = get_data_dir()
    max_filename_length = 255
    if len(filename + ".md") > max_filename_length:
        filename = filename[0 : max_filename_length - 3]
    path_to_md_file = data_dir / path.strip("/") / f"{filename}.md"
    with open(path_to_md_file, "w", encoding="utf-8") as file:
        file.write(contents)

    return path_to_md_file


def get_item(dataobj_id):
    """Returns a Post object with the given dataobjs' attributes"""
    file = get_by_id(dataobj_id)
    if file:
        data = frontmatter.load(file)
        data["fullpath"] = str(file)
        data["dir"] = str(file.parent.relative_to(get_data_dir()))
        # replace . for root items to ''
        if data["dir"] == ".":
            data["dir"] = ""
        return data
    return None


def move_item(dataobj_id, new_path):
    """Move dataobj of given id to new_path"""
    file = get_by_id(dataobj_id)
    data_dir = get_data_dir()
    out_dir = (data_dir / new_path).resolve()
    if not file:
        raise FileNotFoundError
    if (out_dir / file.parts[-1]).exists():
        raise FileExistsError
    elif is_relative_to(out_dir, data_dir) and out_dir.exists():  # check file isn't
        return shutil.move(str(file), f"{get_data_dir()}/{new_path}/")
    return False


def rename_folder(old_path, new_name):
    data_dir = get_data_dir()
    curr_dir = (data_dir / old_path).resolve()
    suggested_renaming = (curr_dir.parent / new_name).resolve()
    if (
        not is_relative_to(suggested_renaming, data_dir)
        or curr_dir == data_dir
        or suggested_renaming == data_dir
        or not is_relative_to(curr_dir, data_dir)
        or not suggested_renaming.parent.is_dir()
    ):
        return False  # invalid inputs
    if not curr_dir.is_dir():
        raise FileNotFoundError
    if suggested_renaming.exists():
        raise FileExistsError
    curr_dir.rename(suggested_renaming)
    return str(suggested_renaming.relative_to(data_dir))


def delete_item(dataobj_id):
    """Delete dataobj of given id"""
    file = get_by_id(dataobj_id)
    remove_from_index(dataobj_id)
    if file:
        Path(file).unlink()


def update_item_md(dataobj_id, new_content):
    """
    Given an object id, this method overwrites the inner
    content of the post with `new_content`.

    This means that it won't change the frontmatter (eg tags, id, title)
    but it can change the file content.

    For example:

    If we have a dataobj like this:

    ```md
    ---
    id: 1
    title: Note
    ---

    # This is random
    ```

    Calling `update_item(1, "# This is specific")` will turn it into:


    ```md
    ---
    id: 1 # unchanged
    title: Note
    ---

    # This is specific
    ```
    """

    from archivy.models import DataObj

    filename = get_by_id(dataobj_id)
    dataobj = frontmatter.load(filename)
    dataobj.content = new_content
    md = frontmatter.dumps(dataobj)
    with open(filename, "w", encoding="utf-8") as f:
        f.write(md)

    converted_dataobj = DataObj.from_md(md)
    converted_dataobj.fullpath = str(
        filename.relative_to(current_app.config["USER_DIR"])
    )
    converted_dataobj.index()
    current_app.config["HOOKS"].on_edit(converted_dataobj)


def update_item_frontmatter(dataobj_id, new_frontmatter):
    """
    Given an object id, this method overwrites the front matter
    of the post with `new_frontmatter`.

    ---
    date: Str
    id: Str
    path: Str
    tags: List[Str]
    title: Str
    type: note/bookmark
    ---
    """

    from archivy.models import DataObj

    filename = get_by_id(dataobj_id)
    dataobj = frontmatter.load(filename)
    for key in list(new_frontmatter):
        dataobj[key] = new_frontmatter[key]
    md = frontmatter.dumps(dataobj)
    with open(filename, "w", encoding="utf-8") as f:
        f.write(md)

    converted_dataobj = DataObj.from_md(md)
    converted_dataobj.fullpath = str(
        filename.relative_to(current_app.config["USER_DIR"])
    )
    converted_dataobj.index()
    current_app.config["HOOKS"].on_edit(converted_dataobj)


def get_dirs():
    """Gets all dir names where dataobjs are stored"""
    # join glob matchers
    dirnames = [
        str(dir_path.relative_to(get_data_dir()))
        for dir_path in get_data_dir().rglob("*")
        if dir_path.is_dir()
    ]

    return dirnames


def create_dir(name):
    """Create dir of given name"""
    root_dir = get_data_dir()
    new_path = root_dir / name.strip("/")
    if is_relative_to(new_path, root_dir):
        new_path.mkdir(parents=True, exist_ok=True)
        return str(new_path.relative_to(root_dir))
    return False


def delete_dir(name):
    """Deletes dir of given name"""
    root_dir = get_data_dir()
    target_dir = root_dir / name
    if not is_relative_to(target_dir, root_dir) or target_dir == root_dir:
        return False
    try:
        shutil.rmtree(target_dir)
        return True
    except FileNotFoundError:
        return False


def format_file(path: str):
    """
    Converts normal md of file at `path` to formatted archivy markdown file, with yaml front matter
    and a filename of format "{id}-{old_filename}.md"
    """

    from archivy.models import DataObj

    data_dir = get_data_dir()
    path = Path(path)
    if not path.exists():
        return

    if path.is_dir():
        for filename in path.iterdir():
            format_file(filename)

    else:
        new_file = path.open("r", encoding="utf-8")
        file_contents = new_file.read()
        new_file.close()
        try:
            # get relative path of object in `data` dir
            datapath = path.parent.resolve().relative_to(data_dir)
        except ValueError:
            datapath = Path()

        note_dataobj = {
            "title": path.name.replace(".md", ""),
            "content": file_contents,
            "type": "note",
            "path": str(datapath),
        }

        dataobj = DataObj(**note_dataobj)
        dataobj.insert()

        path.unlink()
        current_app.logger.info(
            f"Formatted and moved {str(datapath / path.name)} to {dataobj.fullpath}"
        )


def unformat_file(path: str, out_dir: str):
    """
    Converts normal md of file at `path` to formatted archivy markdown file, with yaml front matter
    and a filename of format "{id}-{old_filename}.md"
    """

    data_dir = get_data_dir()
    path = Path(path)
    out_dir = Path(out_dir)
    if not path.exists() and out_dir.exists() and out_dir.is_dir():
        return

    if path.is_dir():
        path.mkdir(exist_ok=True)
        for filename in path.iterdir():
            unformat_file(filename, str(out_dir))

    else:
        dataobj = frontmatter.load(str(path))

        try:
            # get relative path of object in `data` dir
            datapath = path.parent.resolve().relative_to(data_dir)
        except ValueError:
            datapath = Path()

        # create subdir if doesn't exist
        (out_dir / datapath).mkdir(exist_ok=True)
        new_path = out_dir / datapath / f"{dataobj.metadata['title']}.md"
        with new_path.open("w") as f:
            f.write(dataobj.content)

        current_app.logger.info(
            f"Unformatted and moved {str(path)} to {str(new_path.resolve())}"
        )
        path.unlink()


def open_file(path):
    """Cross platform way of opening file on user's computer"""
    if platform.system() == "Windows":
        os.startfile(path)
    elif platform.system() == "Darwin":
        subprocess.Popen(["open", path])
    else:
        subprocess.Popen(["xdg-open", path])


def valid_image_filename(filename):
    ALLOWED_EXTENSIONS = ["jpg", "png", "gif", "jpeg"]
    return "." in filename and filename.rsplit(".", 1)[1] in ALLOWED_EXTENSIONS


def save_image(image: FileStorage):
    """
    Saves image to USER_DATA_DIR

    Returns: filename where image has been saved.
    """
    base_path = Path(current_app.config["USER_DIR"]) / "images"
    fileparts = image.filename.rsplit(".", 1)
    sanitized_filename = secure_filename(fileparts[0])
    dest_path = base_path / f"{sanitized_filename}.{fileparts[1]}"
    i = 1
    while dest_path.exists():
        dest_path = base_path / f"{sanitized_filename}-{i}.{fileparts[1]}"
        i += 1
    image.save(str(dest_path))
    return dest_path.parts[-1]


def image_exists(filename: str):
    sanitized = secure_filename(filename)
    images_dir = Path(current_app.config["USER_DIR"]) / "images"
    image_path = images_dir / sanitized
    if image_path.exists() and is_relative_to(image_path, images_dir):
        return str(image_path)
    return 0



import pprint
from pathlib import Path

import click
import pytest

from archivy import click_web, cli
from archivy.click_web.resources import cmd_form, input_fields


@pytest.mark.parametrize(
    "param, command_index, expected",
    [
        (
            click.Argument(
                [
                    "an_argument",
                ]
            ),
            0,
            {
                "checked": "",
                "click_type": "text",
                "help": "",
                "human_readable_name": "AN ARGUMENT",
                "name": "0.0.argument.text.1.text.an-argument",
                "nargs": 1,
                "param": "argument",
                "required": True,
                "type": "text",
                "value": None,
            },
        ),
        (
            click.Argument(
                [
                    "an_argument",
                ],
                nargs=2,
            ),
            1,
            {
                "checked": "",
                "click_type": "text",
                "help": "",
                "human_readable_name": "AN ARGUMENT",
                "name": "1.0.argument.text.2.text.an-argument",
                "nargs": 2,
                "param": "argument",
                "required": True,
                "type": "text",
                "value": None,
            },
        ),
        (
            click.Option(
                [
                    "--an_option",
                ]
            ),
            0,
            {
                "checked": "",
                "click_type": "text",
                "desc": None,
                "help": ("--an_option TEXT", ""),
                "human_readable_name": "an option",
                "name": "0.0.option.text.1.text.--an-option",
                "nargs": 1,
                "param": "option",
                "required": False,
                "type": "text",
                "value": "",
            },
        ),
        (
            click.Option(
                [
                    "--an_option",
                ],
                nargs=2,
            ),
            1,
            {
                "checked": "",
                "click_type": "text",
                "desc": None,
                "help": ("--an_option TEXT...", ""),
                "human_readable_name": "an option",
                "name": "1.0.option.text.2.text.--an-option",
                "nargs": 2,
                "param": "option",
                "required": False,
                "type": "text",
                "value": "",
            },
        ),
        (
            click.Option(
                [
                    "--flag/--no-flag",
                ],
                default=True,
                help="help",
            ),
            3,
            {
                "checked": 'checked="checked"',
                "click_type": "bool_flag",
                "desc": "help",
                "help": ("--flag / --no-flag", "help"),
                "human_readable_name": "flag",
                "name": "3.0.flag.bool_flag.1.checkbox.--flag",
                "nargs": 1,
                "off_flag": "--no-flag",
                "on_flag": "--flag",
                "param": "option",
                "required": False,
                "type": "checkbox",
                "value": "--flag",
            },
        ),
    ],
)
def test_get_input_field(ctx, click_cli, param, expected, command_index):
    res = input_fields.get_input_field(ctx, param, command_index, 0)
    pprint.pprint(res)
    assert res == expected


@pytest.mark.parametrize(
    "param, command_index, expected",
    [
        (
            click.Argument(
                [
                    "an_argument",
                ],
                nargs=-1,
            ),
            0,
            {
                "checked": "",
                "click_type": "text",
                "help": "",
                "human_readable_name": "AN ARGUMENT",
                "name": "0.0.argument.text.-1.text.an-argument",
                "nargs": -1,
                "param": "argument",
                "required": False,
                "type": "text",
                "value": None,
            },
        ),
    ],
)
def test_variadic_arguments(ctx, click_cli, param, expected, command_index):
    res = input_fields.get_input_field(ctx, param, command_index, 0)
    pprint.pprint(res)
    assert res == expected


@pytest.mark.parametrize(
    "param, command_index, expected",
    [
        (
            click.Argument(
                [
                    "a_file_argument",
                ],
                type=click.File("rb"),
            ),
            0,
            {
                "checked": "",
                "click_type": "file[rb]",
                "help": "",
                "human_readable_name": "A FILE ARGUMENT",
                "name": "0.0.argument.file[rb].1.file.a-file-argument",
                "nargs": 1,
                "param": "argument",
                "required": True,
                "type": "file",
                "value": None,
            },
        ),
        (
            click.Option(
                [
                    "--a_file_option",
                ],
                type=click.File("rb"),
            ),
            0,
            {
                "checked": "",
                "click_type": "file[rb]",
                "desc": None,
                "help": ("--a_file_option FILENAME", ""),
                "human_readable_name": "a file option",
                "name": "0.0.option.file[rb].1.file.--a-file-option",
                "nargs": 1,
                "param": "option",
                "required": False,
                "type": "file",
                "value": "",
            },
        ),
    ],
)
def test_get_file_input_field(ctx, click_cli, param, expected, command_index):
    res = input_fields.get_input_field(ctx, param, command_index, 0)
    pprint.pprint(res)
    assert res == expected


@pytest.mark.parametrize(
    "param, command_index, expected",
    [
        (
            click.Argument(
                [
                    "a_file_argument",
                ],
                type=click.File("wb"),
            ),
            0,
            {
                "checked": "",
                "click_type": "file[wb]",
                "help": "",
                "human_readable_name": "A FILE ARGUMENT",
                "name": "0.0.argument.file[wb].1.hidden.a-file-argument",
                "nargs": 1,
                "param": "argument",
                "required": True,
                "type": "hidden",
                "value": None,
            },
        ),
        (
            click.Option(
                [
                    "--a_file_option",
                ],
                type=click.File("wb"),
            ),
            0,
            {
                "checked": "",
                "click_type": "file[wb]",
                "desc": None,
                "help": ("--a_file_option FILENAME", ""),
                "human_readable_name": "a file option",
                "name": "0.0.option.file[wb].1.text.--a-file-option",
                "nargs": 1,
                "param": "option",
                "required": False,
                "type": "text",
                "value": "",
            },
        ),
    ],
)
def test_get_output_file_input_field(ctx, click_cli, param, expected, command_index):
    res = input_fields.get_input_field(ctx, param, command_index, 0)
    pprint.pprint(res)
    assert res == expected



from flask_wtf import FlaskForm
from wtforms import (
    StringField,
    SubmitField,
    SelectField,
    PasswordField,
    HiddenField,
    BooleanField,
    FormField,
)
from wtforms.fields.html5 import IntegerField
from wtforms.validators import DataRequired, URL
from archivy.config import Config


class NewBookmarkForm(FlaskForm):
    url = StringField("url", validators=[DataRequired(), URL()])
    tags = StringField("tags")
    path = SelectField("Folder")
    submit = SubmitField("Save")


class NewNoteForm(FlaskForm):
    title = StringField("title", validators=[DataRequired()])
    tags = StringField("tags")
    path = SelectField("Folder")
    submit = SubmitField("Save")


class NewFolderForm(FlaskForm):
    parent_dir = HiddenField()
    new_dir = StringField("New folder", validators=[DataRequired()])
    submit = SubmitField("Create sub directory")


class MoveItemForm(FlaskForm):
    path = SelectField("Move to")
    submit = SubmitField("✓")


class RenameDirectoryForm(FlaskForm):
    new_name = StringField("New name", validators=[DataRequired()])
    current_path = HiddenField()
    submit = SubmitField("Rename current folder")


class DeleteDataForm(FlaskForm):
    submit = SubmitField("Delete")


class DeleteFolderForm(FlaskForm):
    dir_name = HiddenField(validators=[DataRequired()])


class UserForm(FlaskForm):
    username = StringField("username")
    password = PasswordField("password")
    submit = SubmitField("Submit")


class TitleForm(FlaskForm):
    title = StringField("title")
    submit = SubmitField("✓")


def config_form(current_conf, sub=0, allowed=vars(Config())):
    """
    This function defines a Config form that loads default configuration values and creates inputs for each field option

    - current_conf: object of current configuration that we will use to set the defaults
    - sub: this function is recursive and can return a sub form to represent the nesting of the config.
        Sub is a boolean value indicating whether the current form is nested.

    - allowed represents a dictionary of the keys that are allowed in our current level of nesting.
        It's fetched from the default config.
    """

    class ConfigForm(FlaskForm):
        pass

    def process_conf_value(name, val):
        """
        Create and set different form fields.

        """
        val_type = type(val)
        if not name in allowed:
            return
        if val_type is dict:
            sub_form = config_form(val, 1, allowed[name])
            setattr(ConfigForm, name, FormField(sub_form))
        elif val_type is int:
            setattr(ConfigForm, name, IntegerField(name, default=val))
        elif val_type is str:
            setattr(ConfigForm, name, StringField(name, default=val))
        elif val_type is bool:
            setattr(ConfigForm, name, BooleanField(name, default=val))
        elif val_type is list:
            setattr(ConfigForm, name, StringField(name, default=", ".join(val)))

    for key, val in current_conf.items():
        process_conf_value(key, val)
    if sub:
        return ConfigForm
    else:
        ConfigForm.submit = SubmitField("Submit")
        return ConfigForm()



from pathlib import Path
from os import environ
from pkg_resources import iter_entry_points

import click
from click_plugins import with_plugins
from flask.cli import FlaskGroup, load_dotenv, shell_command

from archivy import app
from archivy.config import Config
from archivy.click_web import create_click_web_app
from archivy.data import open_file, format_file, unformat_file
from archivy.helpers import load_config, write_config
from archivy.models import User, DataObj


def create_app():
    return app


@with_plugins(iter_entry_points("archivy.plugins"))
@click.group(cls=FlaskGroup, create_app=create_app)
def cli():
    pass


@cli.command("init", short_help="Initialise your archivy application")
@click.pass_context
def init(ctx):
    try:
        load_config()
        click.confirm(
            "Config already found. Do you wish to reset it? "
            "Otherwise run `archivy config`",
            abort=True,
        )
    except FileNotFoundError:
        pass

    config = Config()

    click.echo("This is the archivy installation initialization wizard.")
    data_dir = click.prompt(
        "Enter the full path of the " "directory where you'd like us to store data.",
        type=str,
        default=str(Path(".").resolve()),
    )

    desires_search = click.confirm(
        "Would you like to enable search on your knowledge base contents?"
    )

    if desires_search:
        search_engine = click.prompt(
            "Then go to https://archivy.github.io/setup-search/ to see the different backends you can use for search and how you can configure them.",
            type=click.Choice(["elasticsearch", "ripgrep", "cancel"]),
            show_choices=True,
        )
        if search_engine != "cancel":
            config.SEARCH_CONF["enabled"] = 1
            config.SEARCH_CONF["engine"] = search_engine

    create_new_user = click.confirm("Would you like to create a new admin user?")
    if create_new_user:
        username = click.prompt("Username")
        password = click.prompt("Password", hide_input=True, confirmation_prompt=True)
        if not ctx.invoke(create_admin, username=username, password=password):
            return

    config.HOST = click.prompt(
        "Host [localhost (127.0.0.1)]",
        type=str,
        default="127.0.0.1",
        show_default=False,
    )

    config.override({"USER_DIR": data_dir})
    app.config["USER_DIR"] = data_dir

    # create data dir
    (Path(data_dir) / "data").mkdir(exist_ok=True, parents=True)

    write_config(vars(config))
    click.echo(
        "Config successfully created at "
        + str((Path(app.config["INTERNAL_DIR"]) / "config.yml").resolve())
    )


@cli.command("config", short_help="Open archivy config.")
def config():
    open_file(str(Path(app.config["INTERNAL_DIR"]) / "config.yml"))


@cli.command("hooks", short_help="Creates hook file if it is not setup and opens it.")
def hooks():
    hook_path = Path(app.config["USER_DIR"]) / "hooks.py"
    if not hook_path.exists():
        with hook_path.open("w") as f:
            f.write(
                "from archivy.config import BaseHooks\n"
                "class Hooks(BaseHooks):\n"
                "   # see available hooks at https://archivy.github.io/reference/hooks/\n"
                "   def on_dataobj_create(self, dataobj): # for example\n"
                "       pass"
            )
    open_file(hook_path)


@cli.command("run", short_help="Runs archivy web application")
def run():
    click.echo("Running archivy...")
    load_dotenv()
    environ["FLASK_RUN_FROM_CLI"] = "false"
    app_with_cli = create_click_web_app(click, cli, app)
    app_with_cli.run(host=app.config["HOST"], port=app.config["PORT"])


@cli.command(short_help="Creates a new admin user")
@click.argument("username")
@click.password_option()
def create_admin(username, password):
    if len(password) < 8:
        click.echo("Password length too short")
        return False
    else:
        user = User(username=username, password=password, is_admin=True)
        if user.insert():
            click.echo(f"User {username} successfully created.")
            return True
        else:
            click.echo("User with given username already exists.")
            return False


@cli.command(short_help="Format normal markdown files for archivy.")
@click.argument("filenames", type=click.Path(exists=True), nargs=-1)
def format(filenames):
    for path in filenames:
        format_file(path)


@cli.command(short_help="Convert archivy-formatted files back to normal markdown.")
@click.argument("filenames", type=click.Path(exists=True), nargs=-1)
@click.argument("output_dir", type=click.Path(exists=True, file_okay=False))
def unformat(filenames, output_dir):
    for path in filenames:
        unformat_file(path, output_dir)


@cli.command(short_help="Sync content to Elasticsearch")
def index():
    data_dir = Path(app.config["USER_DIR"]) / "data"

    if not app.config["SEARCH_CONF"]["enabled"]:
        click.echo("Search must be enabled for this command.")
        return

    for filename in data_dir.rglob("*.md"):
        cur_file = open(filename)
        dataobj = DataObj.from_md(cur_file.read())
        cur_file.close()

        if dataobj.index():
            click.echo(f"Indexed {dataobj.title}...")
        else:
            click.echo(f"Failed to index {dataobj.title}")


if __name__ == "__main__":
    cli()



from pathlib import Path
from os.path import sep
from pkg_resources import require

import frontmatter
from flask import (
    render_template,
    flash,
    redirect,
    request,
    url_for,
    send_file,
    send_from_directory,
)
from flask_login import login_user, current_user, logout_user
from tinydb import Query
from werkzeug.security import check_password_hash, generate_password_hash

from archivy.models import DataObj, User
from archivy import data, app, forms
from archivy.helpers import get_db, write_config
from archivy.search import search
from archivy.config import Config


@app.context_processor
def pass_defaults():
    dataobjs = data.get_items()
    version = require("archivy")[0].version
    SEP = sep
    # check windows parsing for js (https://github.com/Uzay-G/archivy/issues/115)
    if SEP == "\\":
        SEP += "\\"
    return dict(dataobjs=dataobjs, SEP=SEP, version=version)


@app.before_request
def check_perms():
    allowed_path = (
        request.path.startswith("/login")
        or request.path.startswith("/static")
        or request.path.startswith("/api/login")
    )
    if not current_user.is_authenticated and not allowed_path:
        return redirect(url_for("login", next=request.path))
    return


@app.route("/")
@app.route("/index")
def index():
    path = request.args.get("path", "").lstrip("/")
    try:
        files = data.get_items(path=path)
    except FileNotFoundError:
        flash("Directory does not exist.", "error")
        return redirect("/")

    return render_template(
        "home.html",
        title=path or "root",
        search_enabled=app.config["SEARCH_CONF"]["enabled"],
        dir=files,
        current_path=path,
        new_folder_form=forms.NewFolderForm(),
        delete_form=forms.DeleteFolderForm(),
        rename_form=forms.RenameDirectoryForm(),
        view_only=0,
        search_engine=app.config["SEARCH_CONF"]["engine"],
    )


# TODO: refactor two following methods
@app.route("/bookmarks/new", methods=["GET", "POST"])
def new_bookmark():
    default_dir = app.config.get("DEFAULT_BOOKMARKS_DIR", "root directory")
    form = forms.NewBookmarkForm(path=default_dir)
    form.path.choices = [("", "root directory")] + [
        (pathname, pathname) for pathname in data.get_dirs()
    ]
    if form.validate_on_submit():
        path = form.path.data
        tags = form.tags.data.split(",") if form.tags.data != "" else []
        bookmark = DataObj(url=form.url.data, tags=tags, path=path, type="bookmark")
        bookmark.process_bookmark_url()
        bookmark_id = bookmark.insert()
        if bookmark_id:
            flash("Bookmark Saved!", "success")
            return redirect(f"/dataobj/{bookmark_id}")
    # for bookmarklet
    form.url.data = request.args.get("url", "")
    path = request.args.get("path", default_dir).strip("/")
    # handle empty argument
    form.path.data = path
    return render_template("dataobjs/new.html", title="New Bookmark", form=form)


@app.route("/notes/new", methods=["GET", "POST"])
def new_note():
    form = forms.NewNoteForm()
    default_dir = "root directory"
    form.path.choices = [("", default_dir)] + [
        (pathname, pathname) for pathname in data.get_dirs()
    ]
    if form.validate_on_submit():
        path = form.path.data
        tags = form.tags.data.split(",") if form.tags.data != "" else []
        note = DataObj(title=form.title.data, tags=tags, path=path, type="note")
        note_id = note.insert()
        if note_id:
            flash("Note Saved!", "success")
            return redirect(f"/dataobj/{note_id}")
    path = request.args.get("path", default_dir).strip("/")
    # handle empty argument
    form.path.data = path
    return render_template("/dataobjs/new.html", title="New Note", form=form)


@app.route("/dataobj/<dataobj_id>")
def show_dataobj(dataobj_id):
    dataobj = data.get_item(dataobj_id)

    if not dataobj:
        flash("Data could not be found!", "error")
        return redirect("/")

    if request.args.get("raw") == "1":
        return frontmatter.dumps(dataobj)

    backlinks = []
    if app.config["SEARCH_CONF"]["enabled"]:
        if app.config["SEARCH_CONF"]["engine"] == "ripgrep":
            query = f"/{dataobj_id}\)]]"  # escape query for rg
        else:
            query = f"/{dataobj_id})]]"
        incoming_links = search(query, strict=True)
        if incoming_links:
            for hit in incoming_links:
                if hit["id"] != dataobj_id:
                    backlinks.append({"title": hit["title"], "id": hit["id"]})

    # Form for moving data into another folder
    move_form = forms.MoveItemForm()
    move_form.path.choices = [("", "root directory")] + [
        (pathname, pathname) for pathname in data.get_dirs()
    ]

    post_title_form = forms.TitleForm()
    post_title_form.title.data = dataobj["title"]

    return render_template(
        "dataobjs/show.html",
        title=dataobj["title"],
        dataobj=dataobj,
        backlinks=backlinks,
        current_path=dataobj["dir"],
        form=forms.DeleteDataForm(),
        view_only=0,
        search_enabled=app.config["SEARCH_CONF"]["enabled"],
        post_title_form=post_title_form,
        move_form=move_form,
    )


@app.route("/dataobj/move/<dataobj_id>", methods=["POST"])
def move_item(dataobj_id):
    form = forms.MoveItemForm()
    out_dir = form.path.data if form.path.data != "" else "root directory"
    if form.path.data == None:
        flash("No path specified.")
        return redirect(f"/dataobj/{dataobj_id}")
    try:
        if data.move_item(dataobj_id, form.path.data):
            flash(f"Data successfully moved to {out_dir}.", "success")
            return redirect(f"/dataobj/{dataobj_id}")
        else:
            flash(f"Data could not be moved to {out_dir}.", "error")
            return redirect(f"/dataobj/{dataobj_id}")
    except FileNotFoundError:
        flash("Data not found.", "error")
        return redirect("/")
    except FileExistsError:
        flash("Data already in target directory.", "error")
        return redirect(f"/dataobj/{dataobj_id}")


@app.route("/dataobj/delete/<dataobj_id>", methods=["DELETE", "GET"])
def delete_data(dataobj_id):
    try:
        data.delete_item(dataobj_id)
    except BaseException:
        flash("Data could not be found!", "error")
        return redirect("/")
    flash("Data deleted!", "success")
    return redirect("/")


@app.route("/login", methods=["GET", "POST"])
def login():
    form = forms.UserForm()
    if form.validate_on_submit():
        db = get_db()
        user = db.search(
            (Query().username == form.username.data) & (Query().type == "user")
        )

        if user and check_password_hash(user[0]["hashed_password"], form.password.data):
            user = User.from_db(user[0])
            login_user(user, remember=True)
            flash("Login successful!", "success")

            next_url = request.args.get("next")
            return redirect(next_url or "/")

        flash("Invalid credentials", "error")
        return redirect("/login")
    return render_template("users/login.html", form=form, title="Login")


@app.route("/logout", methods=["DELETE", "GET"])
def logout():
    logout_user()
    flash("Logged out successfully", "success")
    return redirect("/")


@app.route("/user/edit", methods=["GET", "POST"])
def edit_user():
    form = forms.UserForm()
    if form.validate_on_submit():
        db = get_db()
        db.update(
            {
                "username": form.username.data,
                "hashed_password": generate_password_hash(form.password.data),
            },
            doc_ids=[current_user.id],
        )
        flash("Information saved!", "success")
        return redirect("/")
    form.username.data = current_user.username
    return render_template("users/edit.html", form=form, title="Edit Profile")


@app.route("/folders/create", methods=["POST"])
def create_folder():
    form = forms.NewFolderForm()
    if form.validate_on_submit():
        path = Path(form.parent_dir.data.strip("/")) / form.new_dir.data
        new_path = data.create_dir(str(path))
        flash("Folder successfully created.", "success")
        return redirect(f"/?path={new_path}")
    flash("Could not create folder.", "error")
    return redirect(request.referrer or "/")


@app.route("/folders/delete", methods=["POST"])
def delete_folder():
    form = forms.DeleteFolderForm()
    if form.validate_on_submit():
        if data.delete_dir(form.dir_name.data):
            flash("Folder successfully deleted.", "success")
            return redirect("/")
        else:
            flash("Folder not found.", "error")
            return redirect(request.referrer or "/", 404)
    flash("Could not delete folder.", "error")
    return redirect(request.referrer or "/")


@app.route("/folders/rename", methods=["POST"])
def rename_folder():
    form = forms.RenameDirectoryForm()
    if form.validate_on_submit():
        try:
            new_path = data.rename_folder(form.current_path.data, form.new_name.data)
            if not new_path:
                flash("Invalid input.", "error")
            else:
                flash("Renamed successfully.", "success")
                return redirect(f"/?path={new_path}")
        except FileNotFoundError:
            flash("Directory not found.", "error")
        except FileExistsError:
            flash("Target directory exists.", "error")
    return redirect("/")


@app.route("/bookmarklet")
def bookmarklet():
    return render_template("bookmarklet.html", title="Bookmarklet")


@app.route("/images/<filename>")
def serve_image(filename):
    if filename and data.valid_image_filename(filename):
        image_path = data.image_exists(filename)
        if image_path:
            return send_file(image_path)
        else:
            return "Image not found", 404
    else:
        return "Invalid file request", 413


@app.route("/static/custom.css")
def custom_css():
    if not app.config["THEME_CONF"].get("use_custom_css", False):
        return ""
    return send_from_directory(
        Path(app.config["USER_DIR"]) / "css",
        app.config["THEME_CONF"]["custom_css_file"],
    )


@app.route("/config", methods=["GET", "POST"])
def config():
    """
    Web View to edit and update configuration.
    """

    def update_config_value(key, val, dictionary):
        if key != "SECRET_KEY":
            if type(val) is dict:
                for k, v in val.items():
                    update_config_value(k, v, dictionary[key])
            else:
                dictionary[key] = val

    form = forms.config_form(app.config)
    default = vars(Config())
    if form.validate_on_submit():
        changed_config = Config()
        changed_config.override(form.data)
        for k, v in vars(changed_config).items():
            # propagate changes to configuration
            update_config_value(k, v, app.config)
        write_config(vars(changed_config))  # save to filesystem config
        flash("Config successfully updated.", "success")
    elif request.method == "POST":
        flash("Could not update config.", "error")
    return render_template(
        "config.html", conf=form, default=default, title="Edit Config"
    )



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", encoding="utf-8") as f:
    all_reqs = f.read().split("\n")
    install_requires = [
        x.strip()
        for x in all_reqs
        if not x.startswith("#") and not x.startswith("-e git")
    ]

setuptools.setup(
    name="archivy",
    version="1.5.0",
    author="Uzay-G",
    author_email="uzgirit@gmail.com",
    description=(
        "Minimalist knowledge base focused on digital preservation"
        " and building your second brain."
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/archivy/archivy",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
    entry_points={
        "console_scripts": [
            "archivy = archivy.cli:cli",
        ]
    },
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    python_requires=">=3.6",
)



from pathlib import Path
import sys

import elasticsearch
import yaml
from elasticsearch import Elasticsearch
from flask import current_app, g
from tinydb import TinyDB, Query, operations

from archivy.config import BaseHooks, Config


def load_config(path=""):
    """Loads `config.yml` file safely and deserializes it to a python dict."""
    path = path or current_app.config["INTERNAL_DIR"]
    with (Path(path) / "config.yml").open() as f:
        return yaml.load(f.read(), Loader=yaml.SafeLoader)


def config_diff(curr_key, curr_val, parent_dict, defaults):
    """
    This function diffs the user config with the defaults to only save what is actually different.

    Returns 1 if the current element or its nested elements are different and have been preserved.
    """
    if type(curr_val) is dict:
        # the any call here diffs all nested children of the current dict and returns whether any have modifications
        if not any(
            [
                config_diff(k, v, curr_val, defaults[curr_key])
                for k, v in list(curr_val.items())
            ]
        ):
            parent_dict.pop(curr_key)
            return 0
    else:
        if defaults[curr_key] == curr_val:
            parent_dict.pop(curr_key)
            return 0
    return 1


def write_config(config: dict):
    """
    Writes a new config dict to a `config.yml` file that will override defaults.
    Compares user config with defaults to only save changes.
    """
    defaults = vars(Config())
    for k, v in list(config.items()):
        if k != "SECRET_KEY":
            config_diff(k, v, config, defaults)
    with (Path(current_app.config["INTERNAL_DIR"]) / "config.yml").open("w") as f:
        yaml.dump(config, f)


def load_hooks():
    try:
        user_hooks = (Path(current_app.config["USER_DIR"]) / "hooks.py").open()
    except FileNotFoundError:
        return BaseHooks()

    user_locals = {}
    exec(user_hooks.read(), globals(), user_locals)
    user_hooks.close()
    return user_locals.get("Hooks", BaseHooks)()


def load_scraper():
    try:
        user_scraping = (Path(current_app.config["USER_DIR"]) / "scraping.py").open()
    except FileNotFoundError:
        return {}
    user_locals = {}
    exec(user_scraping.read(), globals(), user_locals)
    user_scraping.close()
    return user_locals.get("PATTERNS", {})


def get_db(force_reconnect=False):
    """
    Returns the database object that you can use to
    store data persistently
    """
    if "db" not in g or force_reconnect:
        g.db = TinyDB(str(Path(current_app.config["INTERNAL_DIR"]) / "db.json"))

    return g.db


def get_max_id():
    """Returns the current maximum id of dataobjs in the database."""
    db = get_db()
    max_id = db.search(Query().name == "max_id")
    if not max_id:
        db.insert({"name": "max_id", "val": 0})
        return 0
    return max_id[0]["val"]


def set_max_id(val):
    """Sets a new max_id"""
    db = get_db()
    db.update(operations.set("val", val), Query().name == "max_id")


def test_es_connection(es):
    """Tests health and presence of connection to elasticsearch."""
    try:
        health = es.cluster.health()
    except elasticsearch.exceptions.ConnectionError:
        current_app.logger.error(
            "Elasticsearch does not seem to be running on "
            f"{current_app.config['SEARCH_CONF']['url']}. Please start "
            "it, for example with: sudo service elasticsearch restart"
        )
        current_app.logger.error(
            "You can disable Elasticsearch by modifying the `enabled` variable "
            f"in {str(Path(current_app.config['INTERNAL_DIR']) / 'config.yml')}"
        )
        sys.exit(1)

    if health["status"] not in ("yellow", "green"):
        current_app.logger.warning(
            "Elasticsearch reports that it is not working "
            "properly. Search might not work. You can disable "
            "Elasticsearch by setting ELASTICSEARCH_ENABLED to 0."
        )


def get_elastic_client(error_if_invalid=True):
    """Returns the elasticsearch client you can use to search and insert / delete data"""
    if (
        not current_app.config["SEARCH_CONF"]["enabled"]
        or current_app.config["SEARCH_CONF"]["engine"] != "elasticsearch"
    ) and error_if_invalid:
        return None

    auth_setup = (
        current_app.config["SEARCH_CONF"]["es_user"]
        and current_app.config["SEARCH_CONF"]["es_password"]
    )
    if auth_setup:
        es = Elasticsearch(
            current_app.config["SEARCH_CONF"]["url"],
            http_auth=(
                current_app.config["SEARCH_CONF"]["es_user"],
                current_app.config["SEARCH_CONF"]["es_password"],
            ),
        )
    else:
        es = Elasticsearch(current_app.config["SEARCH_CONF"]["url"])
    if error_if_invalid:
        test_es_connection(es)
    else:
        try:
            es.cluster.health()
        except elasticsearch.exceptions.ConnectionError:
            return False
    return es
